""" Compatability fixes to make Python 2.7 look more like Python 3.

The general approach is to code using the common subset offered by 'six'.

The HTTPMessage class has a different interface.  This work-arounds makes the
Python 2.7 look enough like the Python 3 for the Wextracto code to work.
"""

import six

if six.PY2:

    import urllib
    import tarfile
    from httplib import HTTPMessage

    def get_content_subtype(self):
        return self.getsubtype()
    HTTPMessage.get_content_subtype = get_content_subtype

    def get_content_charset(self):
        return self.getparam('charset')
    HTTPMessage.get_content_charset = get_content_charset

    def parse_headers(fp):
        return HTTPMessage(fp, 0)

    def urlquote(name, *args, **kwargs):
        if isinstance(name, unicode):
            name = name.encode('utf-8')
        return urllib.quote(name, *args, **kwargs)

    def xzopen(cls, name, mode="r", fileobj=None, compresslevel=9, **kwargs):
        """Open lzma compressed tar archive name for reading or writing.
           Appending is not allowed.
        """
        if mode not in ("r", "w"):
            raise ValueError("mode must be 'r' or 'w'")

        try:
            from backports import lzma
            lzma.LZMAFile
        except (ImportError, AttributeError):
            raise tarfile.CompressionError("lzma module is not available")

        try:
            fileobj = lzma.LZMAFile(fileobj or name, mode)
        except (OSError, IOError):
            if mode == 'r':
                raise tarfile.ReadError("not an lzma file")
            raise

        try:
            fileobj.peek()
        except (lzma.LZMAError, EOFError):
            raise tarfile.ReadError("not an lzma file")

        try:
            t = cls.taropen(name, mode, fileobj, **kwargs)
        except IOError:
            fileobj.close()
            if mode == 'r':
                raise tarfile.ReadError("not an lzma file")
            raise
        except:
            fileobj.close()
            raise
        t._extfileobj = False
        return t

    tarfile.TarFile.xzopen = classmethod(xzopen)
    tarfile.TarFile.OPEN_METH['xz'] = 'xzopen'

else:

    from http.client import parse_headers  # pragma: no cover
    assert parse_headers                   # pragma: no cover

    from six.moves.urllib_parse import quote as urlquote
    assert urlquote



""" Extracted data values are represented with tab-separated fields.
The right-most field on each line is the value, all preceding fields
are labels that describe the value.
The labels and the value are all JSON encoded.

So for example, a value 9.99 with a labels ``product`` and ``price`` would 
look like::

    "product"\t"price"\t9.99\n

And we could decode this line with the following Python snippet:

.. code-block:: pycon

    >>> import json
    >>> line = '"product"\\t"price"\\t9.99\\n'
    >>> [json.loads(s) for s in line.split('\\t')]
    [u'product', u'price', 9.99]

Using tab-delimiters is convenient for downstream processing using Unix 
command line tools such as :command:`cut` and :command:`grep`.
"""
from __future__ import unicode_literals, print_function
import sys
import json
import logging
from itertools import product
from operator import itemgetter
from bdb import BdbQuit
from six import PY2, binary_type, text_type, reraise
from six.moves import map
from .iterable import walk, flatten, do_not_iter

logger = logging.getLogger(__name__)

TAB = '\t'
NL = '\n'

json_encoder_kwargs = dict(
    separators=(',', ':'),
    default=None,
    sort_keys=True,
    ensure_ascii=False,
)


if PY2:
    json_encoder_kwargs['encoding'] = 'utf-8'


encode_json = json.JSONEncoder(**json_encoder_kwargs).encode


def encode_field(obj):
    try:
        s = encode_json(obj)
        if isinstance(s, binary_type):
            return s.decode('utf-8')
        return s
    except TypeError:
        return '#' + text_type(repr(obj)) + '!'


def should_iter_unless_list(obj):
    # a list is a reasonable value type so don't flatten it
    return (hasattr(obj, '__iter__') and
            not isinstance(obj, do_not_iter + (list,)))


class Value(tuple):

    exit_on_exc = False
    debug_on_exc = False

    value = property(itemgetter(-1))
    labels = property(itemgetter(slice(0, -1)))

    def __new__(cls, value=(None,)):
        if not isinstance(value, tuple):
            value = (value,)
        return super(Value, cls).__new__(cls, value)

    def text(self):
        """ Returns the text this value as a labelled JSON line. """
        iterables = [map(encode_field, flatten(label))
                     for label in self.labels]
        flattened = flatten(self.value, should_iter_unless_list)
        iterables.append(map(encode_field, flattened))
        for fields in product(*iterables):
            yield TAB.join(fields) + NL

    def label(self, *labels):
        """ Adds zero or more labels to this value. """
        return self.__class__(tuple(labels) + self)


def yield_values(extract, *args, **kw):
    """ Yields ``Value`` objects extracted using ``extract``. """
    exc_info = ()

    try:
        returned = extract(*args, **kw)
        for walker in walk(returned, should_iter_unless_list):
            for value in walker:
                yield Value(value)
    except BdbQuit:
        raise
    except Exception as exc:
        exc_info = sys.exc_info()
        yield Value(exc)

    if any(exc_info) and (Value.exit_on_exc or Value.debug_on_exc):
        if Value.debug_on_exc:
            import traceback
            try:
                import ipdb as pdb
            except ImportError:
                import pdb
                assert pdb
            traceback.print_tb(exc_info[2])
            pdb.post_mortem(exc_info[2])
        else:
            reraise(exc_info[0], exc_info[1], exc_info[2])



import codecs
import requests
from six import iteritems
from six import BytesIO
from six.moves.urllib_parse import urljoin
from lxml.html import _nons, HTMLParser
from .py2compat import parse_headers
from .iterable import one
from .http import timeout, readable_from_response, merge_setting
from .etree import get_base_url
from .response import Response


def create_html_parser(headers):

    charset = headers.get_content_charset()
    try:
        if charset and codecs.lookup(charset).name == 'iso8859-1':
            charset = 'windows-1252'
    except LookupError:
        pass

    # if charset is not specified in the Content-Type, this will be
    # None ; encoding=None produces default (ISO 8859-1) behavior.
    return HTMLParser(encoding=charset)


class ParserReadable(object):
    """ Readable that feeds a parser as it is reads. """

    def __init__(self, readable):
        self.readable = readable
        self.lines = []
        self.code = None
        self.headers = None
        self.parser = None
        self.root = None

    @classmethod
    def from_response(cls, response, url, decode_content, context):
        return cls(readable_from_response(response, url,
                                          decode_content=decode_content,
                                          context=context))

    @property
    def name(self):
        return getattr(self.readable, 'name')

    def read(self, size):
        buf = self.readable.read(size)
        if self.parser:
            self.parser.feed(buf)
            if len(buf) < size:
                if self.root is None:
                    self.root = self.parser.close()
                    url = self.headers.get('X-wex-request-url')
                    # this sets the .base_url
                    self.root.getroottree().docinfo.URL = url
        return buf

    def readline(self, *args):
        line = self.readable.readline(*args)
        if not self.lines:
            _, _, self.code, _ = Response.parse_status_line(self.readable,
                                                            line)
        self.lines.append(line)
        if not line.strip():
            self.headers = parse_headers(BytesIO(b''.join(self.lines[1:])))
            if 200 <= self.code < 300:
                self.parser = create_html_parser(self.headers)
        return line

    def close(self):
        self.readable.close()


# just like:
# https://github.com/lxml/lxml/blob/master/src/lxml/html/__init__.py#L1004
# but doesn't ignore <input type="submit" ...> elements
def form_values(self):
    """
    Return a list of tuples of the field values for the form.
    This is suitable to be passed to ``urllib.urlencode()``.
    """
    results = []
    for el in self.inputs:
        name = el.name
        if not name:
            continue
        tag = _nons(el.tag)
        if tag == 'textarea':
            results.append((name, el.value))
        elif tag == 'select':
            value = el.value
            if el.multiple:
                for v in value:
                    results.append((name, v))
            elif value is not None:
                results.append((name, el.value))
        else:
            assert tag == 'input', (
                "Unexpected tag: %r" % el)
            if el.checkable and not el.checked:
                continue
            if el.type in ('image', 'reset'):
                continue
            value = el.value
            if value is not None:
                results.append((name, el.value))
    return results


def submit_form(url, method, session=None, **kw):

    if session is None:
        session = requests.Session()
        session.stream = True

    decode_content = kw.get('decode_content', True)
    proxies = kw.get('proxies', None)
    context = kw.get('context', {})
    headers = merge_setting(method.args.get('headers'), kw.get('headers'))

    response = session.request(
        'get',
        url,
        allow_redirects=False,
        cookies=method.args.get('cookies', None),
        data=None,
        headers=headers,
        params=method.args.get('params', None),
        proxies=proxies,
        timeout=timeout,
    )
    readable = ParserReadable.from_response(response, url,
                                            decode_content=decode_content,
                                            context=context)
    yield readable

    redirects = session.resolve_redirects(response,
                                          response.request,
                                          proxies=proxies,
                                          stream=True,
                                          timeout=timeout)

    for response in redirects:
        readable = ParserReadable.from_response(response, url,
                                                decode_content=decode_content,
                                                context=context)
        yield readable

    if readable.root is None:
        return

    form_css_selector, values = one(iteritems(method.args))
    form = one(readable.root.cssselect(form_css_selector))

    if isinstance(values, dict):
        values = values.items()

    for name, value in values:
        if name in form.inputs:
            input = form.inputs[name]
        else:
            input = one(form.cssselect(name))
        if hasattr(input, 'add'):
            input.add(value)
        else:
            input.value = value

    base_url = get_base_url(form)
    form_action_url = urljoin(base_url, form.get('action', ''))

    form_method = form.method.upper()
    if form_method in ('POST', 'PUT'):
        # this implies 'application/x-www-form-urlencoded'
        data = form_values(form)
        params = None
    else:
        data = None
        params = form_values(form)

    response = session.request(
        form_method,
        form_action_url,
        params=params,
        allow_redirects=False,
        cookies=method.args.get('cookies', None),
        data=data,
        headers=headers,
        proxies=proxies,
        timeout=timeout,
    )
    yield readable_from_response(response, url,
                                 decode_content=decode_content,
                                 context=context)

    redirects = session.resolve_redirects(response,
                                          response.request,
                                          proxies=proxies,
                                          stream=True,
                                          timeout=timeout)
    for redirect in redirects:
        yield readable_from_response(redirect, url,
                                     decode_content=decode_content,
                                     context=context)
